from typing import Dict, Any, Optional, List, Tuple
import logging

from pydantic import BaseModel

LOGGER = logging.getLogger(__name__)

HANDLER_TYPES = {
    "PostgreSQL": "pg",
    "Microsoft SQL Server": "mssql",
    "Apache Hive": "hive",
    "Apache Impala": "impala",
    "Apache HDFS": "hdfs",
    "Azure Blob Storage": "azureblob",
    "Azure SQL Data Warehouse": "asdw",
    "Netezza": "netezza",
    "MariaDB": "mariadb",
    "DB2": "db2",
    "Oracle": "oracle",
    "MySQL": "mysql",
    "Elastic": "elastic",
    "Teradata": "teradata",
    "Greenplum": "greenplum",
    "Redshift": "redshift",
    "Amazon S3": "s3",
    "FTP": "ftp",
    "Persisted": "persisted",
    "Custom": "custom",
    "MEMSQL": "memsql",
    "Presto": "presto",
    "Amazon Athena": "athena",
    "Vertica": "vertica",
}

PREFIX_MAP = {
    "PostgreSQL": "pg",
    "Redshift": "rs",
    "Amazon S3": "s3",
    "Amazon Athena": "ath",
}
MAX_IMMUTA_NAME_LIMIT = 55
MAX_POSTGRES_NAME_LIMIT = 63


def blob_handler_type(handler_type: str) -> str:
    return HANDLER_TYPES.get(handler_type, handler_type)


def make_immuta_table_name(
    handler_type: str, schema: str, table: str, user_prefix: Optional[str]
) -> str:
    """
    Returns a table name that's guaranteed to be unique and within the max char limit (55)
    """
    table_name = ""
    if user_prefix:
        table_name = f"{user_prefix}_"
    table_name += f"{PREFIX_MAP[handler_type]}_{schema}_{table}"
    if table_name is None:
        return None
    if len(table_name) <= MAX_IMMUTA_NAME_LIMIT:
        return table_name
    import hashlib

    return (
        table_name[: MAX_IMMUTA_NAME_LIMIT - 8]
        + hashlib.md5(table_name.encode()).hexdigest()[:8]
    )


def make_postgres_table_name(
    handler_type: str, schema: str, table: str, user_prefix: Optional[str]
) -> str:
    """
    Returns table name that has a shortened prefix and conforms to the Postgres max char limit (63)
    """
    table_name = ""
    if user_prefix:
        table_name = f"{user_prefix}_"
    table_name += f"{PREFIX_MAP[handler_type]}_{schema}_{table}"
    if len(table_name) < MAX_POSTGRES_NAME_LIMIT:
        return table_name
    trunc_table_name = table_name[:MAX_POSTGRES_NAME_LIMIT]
    LOGGER.warning(
        "Postgres table name too long! Table %s truncated to %s",
        table_name,
        trunc_table_name,
    )
    return trunc_table_name


class BlobHandler(BaseModel):
    scheme: str = ""
    url: str = ""


class ProfileGroup(BaseModel):
    profiles: List[str] = []
    groups: List[str] = []


class DataSource(BaseModel):
    # Describes the type of underlying blob handler that will be used
    # with this Data Source, e.g. Custom, MS SQL.
    blobHandlerType: str
    private: Optional[bool] = None
    # organization: str
    # What to call the data source in Immuta
    name: Optional[str] = None
    # A list of full URLs providing the locations of all Blob Store handlers
    # to use with this Data Source.
    # Unnecessary when creating data source
    blobHandler: Optional[BlobHandler] = BlobHandler()
    # Users and Groups that should be added as owners to this Data Source.
    # Profiles must be a list of profile ID's and groups must be a list of group ids.
    owner: ProfileGroup = ProfileGroup()
    expert: ProfileGroup = ProfileGroup()
    ingest: ProfileGroup = ProfileGroup()
    sqlTableName: Optional[str] = None
    # The category of the Data Source
    # category: str = ""
    description: str = ""
    hasSamples: bool = False
    recordFormat: str = "json"
    # The type of Data Source, whether it is ingested (metadata will exist in Immuta)
    # or queryable (metadata is dynamically queried)
    type: str = "queryable"
    useDatesAsDirectory: bool = False


class DataSourceColumn(BaseModel):
    name: str
    dataType: str
    remoteType: str
    nullable: bool
    tags: List[Dict[str, str]] = []


class HandlerMetadata(BaseModel):
    database: str
    username: str
    password: str
    table: str
    fh_schema: str
    # The length in seconds that data from this handler can be cached.
    staleDataTolerance: int
    # Don't know what this is. . .
    # blobId: List[str]
    # Table name that is exposed by Immuta
    bodataTableName: str = ""
    # The name of the Data Source to which this handler corresponds
    dataSourceName: str = ""
    columns: Optional[List[DataSourceColumn]] = None
    format: str = "json"
    isChildDataSource: bool = False
    ssl: bool = True

    class Config:
        fields = {"fh_schema": "schema"}


class AthenaHandlerMetadata(HandlerMetadata):
    queryResultLocationBucket: str
    queryResultLocationDirectory: str
    region: str
    authenticationMethod: str = "accessKey"


class PostgresHandlerMetadata(HandlerMetadata):
    hostname: str
    port: int = 5432


class Handler(BaseModel):
    metadata: HandlerMetadata


class DataSourceDictionary(BaseModel):
    id: int
    dataSource: int
    metadata: List[DataSourceColumn]
    types: List[str]
    createdAt: str = ""
    updatedAt: str = ""


HANDLER_TO_METADATA_CLASS = {
    "PostgreSQL": PostgresHandlerMetadata,
    "Amazon Athena": AthenaHandlerMetadata,
    "Redshift": PostgresHandlerMetadata,
}


def make_bulk_create_objects(
    config: Dict[str, Any],
    schema: str,
    tables: List[str],
    user_prefix: Optional[str] = None,
) -> Tuple[DataSource, List[Handler]]:
    """
    Returns a (data source, metadata) tuple containing relevant details to bulk create new data
    sources in Immuta from the source schema
    """
    handlers = []
    for table in tables:
        external_table_name = make_postgres_table_name(
            config["handler_type"], schema, table, user_prefix
        )
        immuta_table_name = make_immuta_table_name(
            config["handler_type"], schema, table, user_prefix
        )

        handler = make_handler_metadata(
            table=table,
            schema=schema,
            config=config,
            bodataTableName=external_table_name,
            dataSourceName=immuta_table_name,
        )
        handlers.append(handler)

    ds = DataSource(
        blobHandlerType=config["handler_type"], recordFormat="json", type="queryable"
    )
    return (ds, handlers)


def to_immuta_objects(
    config: Dict[str, Any],
    schema: str,
    table: str,
    columns: List[DataSourceColumn],
    user_prefix: Optional[str] = None,
) -> Tuple[DataSource, Handler]:
    """
    Returns a tuple containing relevant details to create a new data source
    in Immuta from the source schema
    """
    external_table_name = make_postgres_table_name(
        handler_type=config["handler_type"],
        schema=schema,
        table=table,
        user_prefix=user_prefix,
    )
    immuta_table_name = make_immuta_table_name(
        config["handler_type"], schema, table, user_prefix
    )
    handler = make_handler_metadata(
        table=table,
        schema=schema,
        config=config,
        columns=columns,
        bodataTableName=external_table_name,
        dataSourceName=immuta_table_name,
    )
    ds = DataSource(
        name=immuta_table_name,
        sqlTableName=external_table_name,
        blobHandlerType=config["handler_type"],
        blobHandler=BlobHandler(scheme="https"),
        recordFormat="json",
        type="queryable",
        # category="foo",
        description="bar",
        # owner="foo",
    )
    return (ds, handler)


def make_handler_metadata(
    table: str, schema: str, config: Dict[str, Any], **kwargs
) -> Handler:

    # Ensure that required keys are set
    # Defaulting values in the BaseModel doesn't work
    # since we don't use defaults in this particular instance
    required_args = {
        "format": "json",
        "ssl": True,
        "staleDataTolerance": (24 * 60 * 60),
    }
    for k, v in required_args.items():
        config[k] = config.get(k, v)
    if config["handler_type"] == "Amazon Athena":
        metadata = AthenaHandlerMetadata(
            # Default value but has to exist in the final used dict
            authenticationMethod="accessKey",
            **kwargs,
            **locals(),
            **config,
        )
    elif config["handler_type"] in ["PostgreSQL", "Redshift"]:
        metadata = PostgresHandlerMetadata(
            # Default value but has to exist in the final used dict
            **kwargs,
            **locals(),
            **config,
        )
    handler = Handler(metadata=metadata)
    return handler



from functools import partial
import fnmatch
import os
import logging
import glob
from typing import (
    Callable,
    Dict,
    List,
    Any,
    Union,
    Iterator,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

import click
import yaml
import requests
from toolz.itertoolz import groupby
from toolz.dicttoolz import keyfilter

from fh_immuta_utils.authenticate import retrieve_credentials
from fh_immuta_utils.client import get_client
from fh_immuta_utils.config import parse_config
from fh_immuta_utils.data_source import (
    DataSource,
    Handler,
    to_immuta_objects,
    make_handler_metadata,
    make_bulk_create_objects,
)

if TYPE_CHECKING:
    from immuta_utils.client import ImmutaClient

LOGGER = logging.getLogger(__name__)


@click.command()
@click.option("--config-file", required=True)
@click.option(
    "--glob-prefix",
    default="*.yml",
    help="Glob for specific data source spec files matching given prefix",
)
@click.option("--debug", is_flag=True, default=False, help="Debug logging")
@click.option("--dry-run", is_flag=True, default=False, help="Dry run")
def cli_entrypoint(config_file: str, glob_prefix: str, debug: bool, dry_run: bool):
    main(config_file=config_file, glob_prefix=glob_prefix, debug=debug, dry_run=dry_run)


def main(config_file: str, glob_prefix: str, debug: bool, dry_run: bool) -> bool:

    logging.basicConfig(
        format="[%(name)s][%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=(logging.DEBUG if debug else logging.INFO),
    )

    no_enrollment_errors = True

    config = parse_config(config_file=config_file)
    client = get_client(**config)

    dataset_spec_filepath = os.path.join(
        config["config_root"], "enrolled_datasets", glob_prefix
    )
    LOGGER.debug(f"Globbing for files in {dataset_spec_filepath}")
    for filepath in glob.glob(dataset_spec_filepath):
        LOGGER.info("Processing file: %s", filepath)
        with open(filepath) as handle:
            dataset_spec = yaml.safe_load(handle)
        credentials = retrieve_credentials(dataset_spec["credentials"])
        dataset_spec["username"] = credentials["username"]
        dataset_spec["password"] = credentials["password"]

        schema_table_mapping = get_tables_in_database(client, dataset_spec)

        failed_tables = set()
        data_sources_to_enroll = [
            (dataset_spec["schemas_to_enroll"], data_sources_enroll_iterator),
            (dataset_spec["schemas_to_bulk_enroll"], data_sources_bulk_enroll_iterator),
        ]

        for schemas, enroll_iter in data_sources_to_enroll:
            if not schemas:
                continue
            for schema_object in schemas:
                for (data_source, handler) in enroll_iter(  # type: ignore
                    client=client,
                    schema_table_mapping=schema_table_mapping,
                    schema_obj=schema_object,
                    config=dataset_spec,
                ):
                    LOGGER.debug("Data source: %s", data_source.json())
                    if isinstance(handler, list):
                        LOGGER.debug("Handler[0]: %s", handler[0].json())
                    elif isinstance(handler, Handler):
                        LOGGER.debug("Handler: %s", handler.json())
                    else:
                        raise TypeError(
                            f"Unexpected type for handler; Got: {type(handler)}"
                        )
                    if not dry_run:
                        if not create_data_source(
                            client=client, data_source=data_source, handler=handler
                        ):
                            failed_tables.add(data_source.name)
        if failed_tables:
            no_enrollment_errors = False
            LOGGER.warning("Tables that failed creation:")
            for table in failed_tables:
                LOGGER.warning(table)

    LOGGER.info("Finished enrollment")
    return no_enrollment_errors


def get_tables_in_database(
    client: "ImmutaClient", config: Dict[str, Any]
) -> Dict[str, List[Dict[str, str]]]:
    """ Returns a list of schema_name: [tables...] mapping in the database
    specified by the config """
    # Grab list of all tables in all schemas in the database
    tables_in_database = client.get_table_names(config)
    # Group the tables per schema
    return groupby("tableSchema", tables_in_database)


def data_sources_enroll_iterator(
    client: "ImmutaClient",
    schema_table_mapping: Dict[str, List[Dict[str, str]]],
    schema_obj: Dict[str, str],
    config: Dict[str, Any],
) -> Iterator[Tuple[DataSource, Handler]]:
    LOGGER.info("Processing schema_prefix: %s", schema_obj["schema_prefix"])

    matches_prefix = partial(fnmatch.fnmatch, pat=schema_obj["schema_prefix"])

    for schema, tables in keyfilter(matches_prefix, schema_table_mapping).items():
        LOGGER.info("Processing schema: %s", schema)
        for table in tables:
            if not fnmatch.fnmatch(table["tableName"], schema_obj["table_prefix"]):
                continue
            LOGGER.info("Processing table: %s.%s", schema, table["tableName"])
            handler = make_handler_metadata(
                config=config, table=table["tableName"], schema=schema
            )
            columns = client.get_column_types(
                config=config, data_source_type=config["handler_type"], handler=handler
            )

            data_source, handler = to_immuta_objects(
                schema=schema, table=table["tableName"], columns=columns, config=config
            )
            yield (data_source, handler)


def data_sources_bulk_enroll_iterator(
    client: "ImmutaClient",
    schema_table_mapping: Dict[str, List[Dict[str, str]]],
    schema_obj: Dict[str, str],
    config: Dict[str, Any],
) -> Iterator[Tuple[DataSource, List[Handler]]]:

    LOGGER.info("Processing schema_prefix: %s", schema_obj["schema_prefix"])

    matches_prefix = partial(fnmatch.fnmatch, pat=schema_obj["schema_prefix"])

    for schema, tables in keyfilter(matches_prefix, schema_table_mapping).items():
        LOGGER.info("Bulk creating for all tables in schema %s", schema)
        if len(tables) == 0:
            LOGGER.warning("No tables found for schema: %s", schema)
            continue
        data_source, handlers = make_bulk_create_objects(
            schema=schema,
            tables=[table["tableName"] for table in tables],
            config=config,
            user_prefix=config.get("prefix"),
        )
        yield (data_source, handlers)


def create_data_source(
    client: "ImmutaClient",
    data_source: DataSource,
    handler: Union[Handler, List[Handler]],
) -> bool:
    try:
        result = client.create_data_source(data_source, handler)
        if result:
            LOGGER.info(
                "Created data source %s, id: %d", data_source.name, result["id"]
            )
        return True
    except requests.exceptions.HTTPError as err:
        LOGGER.error("Error creating data source %s: %s", data_source.name, err)
        return False


if __name__ == "__main__":
    cli_entrypoint()



from fh_immuta_utils.client import ImmutaClient

client_config = {"base_url": "...", "api_key": "..."}

if __name__ == "__main__":
    client = ImmutaClient(**client_config)

    # populate
    db_host = "<<redshift host>>"
    db_username = "<<user>>"
    db_password = "<<password>>"
    db_port = 5439  # or a different port
    db_database = "<<database name>>"
    db_use_ssl = True

    # what we're doing here is that we're adding a list of known tables from the same schema.
    #
    remote_schemas_and_tables = {
        "public": [
            "table_1",
            "table_2",
            # ...
        ],
        # more schemas can be appended here
    }

    # do notice, no schema is required:
    handlers = []
    handler_type = "redshift"
    blob_handler_type = handler_type.title()

    for remote_schema in remote_schemas_and_tables.keys():
        for remote_table in remote_schemas_and_tables.get(remote_schema, []):
            handlers.append(
                {
                    "metadata": {
                        "ssl": db_use_ssl,
                        "isChildDataSource": False,
                        "userFiles": [],
                        "port": db_port,
                        "hostname": db_host,
                        "database": db_database,
                        "username": db_username,
                        "password": db_password,
                        "format": "csv",
                        "staleDataTolerance": (7 * 24 * 60 * 60),
                        "bodataTableName": remote_table,
                        "dataSourceName": "{}_{}".format(remote_schema, remote_table),
                        "table": remote_table,
                        "schema": remote_schema,
                    }
                }
            )

    # we end up with a flat list of handlers, no table structure is required:

    # the data source payload is rather counter-intuitive, as it does not play an important role
    # in this type of request, but is required nonetheless:
    data_source = {
        "useDatesAsDirectory": False,
        "private": True,
        "blobHandler": {"scheme": "https", "url": ""},
        "policyHandlerType": "None",
        "blobHandlerType": blob_handler_type,
        "recordFormat": "csv",
        "type": "queryable",
        "hasSamples": False,
    }

    client.create_data_sources(
        handler_type=handler_type, handlers=handlers, data_source=data_source
    )



#!/usr/bin/env python

"""
Script that can be invoked to create, update or delete existing governance policies.
"""

import logging
from typing import Optional, TYPE_CHECKING

from tqdm import tqdm
import click

from fh_immuta_utils.client import get_client
from fh_immuta_utils.config import parse_config
from fh_immuta_utils.tagging import Tagger
from fh_immuta_utils.policy import make_global_subscription_policy

if TYPE_CHECKING:
    from immuta_utils.client import ImmutaClient


@click.command()
@click.option("--config-file", required=True)
@click.option(
    "--search-text",
    help="When deleting, will match all policies that contain this string anywhere in their name. Ignored otherwise.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Log the policies that would be updated without affecting any change",
)
@click.option(
    "--delete", is_flag=True, default=False, help="Delete any matching policies"
)
@click.option("--debug", is_flag=True, default=False, help="Debug logging")
def cli_entrypoint(
    config_file: str, search_text: str, dry_run: bool, delete: bool, debug: bool
):
    return main(
        config_file=config_file,
        search_text=search_text,
        dry_run=dry_run,
        delete=delete,
        debug=debug,
    )


def main(config_file: str, search_text: str, dry_run: bool, delete: bool, debug: bool):
    logging.basicConfig(
        format="[%(name)s][%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=(logging.DEBUG if debug else logging.INFO),
    )
    config = parse_config(config_file=config_file)
    client = get_client(**config)

    if delete:
        if search_text is None:
            logging.error(
                "--delete must be invoked with an explicit value for --search-text"
            )
            return False
        return delete_existing_policies(
            client=client, dry_run=dry_run, search_text=search_text, debug=debug
        )
    else:
        return create_or_update_policies(
            client=client,
            config_root=config["config_root"],
            dry_run=dry_run,
            debug=debug,
        )


def delete_existing_policies(
    client: "ImmutaClient", dry_run: bool, search_text: Optional[str], debug: bool
) -> bool:
    progress_iterator = tqdm(client.get_global_policies(search_text=search_text))

    for policy in progress_iterator:
        progress_iterator.set_description(desc=f"Policy {policy.name}")
        logging.info(f"Deleting policy with name {policy.name}, ID: {policy.id}")
        if not dry_run:
            client.delete_global_policy(id=policy.id)
    logging.info("Fin.")
    return True


def create_or_update_policies(
    client: "ImmutaClient", config_root: str, dry_run: bool, debug: bool
) -> bool:
    logging.info("Gathering existing policies")
    existing_policies = {}

    for policy in client.get_global_policies():
        existing_policies[policy.name] = policy

    logging.debug(f"Existing policies: {existing_policies.keys()}")
    tagger = Tagger(config_root=config_root)
    logging.debug(f"Tag groups: {tagger.tag_groups.keys()}")
    progress_iterator = tqdm(tagger.tag_groups.keys())
    for tag_name in progress_iterator:
        progress_iterator.set_description(desc=f"Tag: {tag_name}")
        iam_groups = tagger.tag_groups[tag_name]
        policy_name = f"{tag_name}_access_policy"
        policy = make_global_subscription_policy(
            policy_name=policy_name,
            tags=[tag_name],
            iam_groups=iam_groups,
            tagger=tagger,
        )
        logging.debug(f"Policy to create/update: {policy.json()}")
        if policy_name in existing_policies.keys():
            policy.id = existing_policies[policy_name].id
            logging.debug(f"Existing policy: {existing_policies[policy_name].json()}")
            if existing_policies[policy_name] == policy:
                logging.info(f"No change for policy {policy_name}. Skipping.")
                continue
            logging.info(f"Updating existing policy with name {policy_name}.")
            if not dry_run:
                client.update_global_policy(
                    policy=policy, id=existing_policies[policy_name].id
                )
        else:
            logging.info(f"Creating new policy with name {policy_name}.")
            if not dry_run:
                client.create_global_policy(policy=policy)
    logging.info("Fin.")
    return True


if __name__ == "__main__":
    cli_entrypoint()



from collections import namedtuple
from unittest import mock
from typing import Dict, List, Any
import pytest

import fh_immuta_utils.tagging as tg
import fh_immuta_utils.data_source as ds

TAG_MAP = {"col_foo": ["foo", "foobar"], "col_bar": ["bar.baz"]}
TAG_GROUPS = {
    "foo": ["group_foo", "group_foobar"],
    "foobar": ["group_foobar"],
    "bar.baz": ["group_baz"],
}


@pytest.fixture
def tagger():
    with mock.patch("fh_immuta_utils.tagging.Tagger.read_configs", return_value=None):
        obj = tg.Tagger(config_root="")
        obj.tag_map = TAG_MAP
        obj.tag_groups = TAG_GROUPS
    return obj


@pytest.mark.parametrize("tag,is_root", [("foo", False), ("bar", True)])
def test_is_root_tag(tagger: tg.Tagger, tag: str, is_root: bool):
    assert tagger.is_root_tag(tag) == is_root


@pytest.mark.parametrize(
    "col,expected", [("col_foo", TAG_MAP["col_foo"]), ("bad_col", [])]
)
def test_get_tags_for_column(tagger: tg.Tagger, col: str, expected: List[str]):
    assert tagger.get_tags_for_column(column_name=col) == expected


def test_tags_to_make(tagger):
    assert list(tagger.tags_to_make()) == [
        ("foo", []),
        ("foobar", []),
        ("bar", ["bar.baz"]),
    ]


TagMsgBody = namedtuple("TagMsgBody", ["root_tag", "children", "expected"])
TAG_MSG_BODY = [
    TagMsgBody(root_tag="foo", children=[], expected={"tags": [{"name": "foo"}]}),
    TagMsgBody(
        root_tag="bar",
        children=["baz"],
        expected={
            "rootTag": {"name": "bar", "deleteHierarchy": False},
            "tags": [{"name": "baz"}],
        },
    ),
    TagMsgBody(
        root_tag="foobar",
        children=["foo", "bar"],
        expected={
            "rootTag": {"name": "foobar", "deleteHierarchy": False},
            "tags": [{"name": "foo"}, {"name": "bar"}],
        },
    ),
]


@pytest.mark.parametrize("root_tag,children,expected", TAG_MSG_BODY)
def test_create_message_body_for_tag_creation(
    tagger: tg.Tagger, root_tag: str, children: List[str], expected: Dict[str, Any]
):
    assert (
        tagger.create_message_body_for_tag_creation(
            root_tag=root_tag, children=children
        )
        == expected
    )


def test_enrich_columns_with_tagging(tagger: tg.Tagger):
    columns = [
        ds.DataSourceColumn(name="col_foo", dataType="", remoteType="", nullable=False),
        ds.DataSourceColumn(name="col_bar", dataType="", remoteType="", nullable=False),
        ds.DataSourceColumn(name="bad_col", dataType="", remoteType="", nullable=False),
    ]
    enriched_cols = tagger.enrich_columns_with_tagging(columns)
    assert len(enriched_cols) == len(columns)
    for col in enriched_cols:
        assert len(col.tags) == len(TAG_MAP.get(col.name, []))



# Copyright 2018 Immuta, Inc. Licensed under the Immuta Software License
# Version 0.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
#
#    http://www.immuta.com/licenses/Immuta_Software_License_0.1.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import sys
from six import string_types

try:
    from logging.config import dictConfig
except ImportError:
    from pip.compat.dictconfig import dictConfig  # type: ignore

logging_config = {
    "version": 1,
    "formatters": {"default": {"format": "%(levelname)s:%(name)s:%(message)s"}},
    "filters": {"limit": {"()": "immuta.log.ReverseLevelFilter"}},
    "handlers": {
        "standard": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "formatter": "default",
            "filters": ["limit"],
        },
        "error": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "ERROR",
        },
    },
    "root": {"level": "INFO", "handlers": ["standard", "error"]},
}


class ReverseLevelFilter(logging.Filter):
    """
    Reverse log level Filter
    Causes only log messages less than or equal to level to be processed
    """

    def __init__(self, level=logging.INFO):
        self.level = level

    def filter(self, record):
        return 1 if record.levelno <= self.level else 0


class LoggingMixin(object):
    """
    Logging Mixin class
    Adds a log property configured with the class's logger
    """

    @property
    def logger_name(self):
        if not hasattr(self, "__logger_name"):
            self.__logger_name = "{0}.{1}".format(
                self.__module__, self.__class__.__name__
            )
        return self.__logger_name

    @property
    def log(self):
        return logging.getLogger(self.logger_name)


def init(level=None, debug=[], log_format=None):
    """
    Initialize logging with the default logging configuration

    debug - package(s) to enable debug logging for
    log_format - format for logs
    """
    # if log_format:
    #     logging_config['formatters']['default']['format'] = log_format
    # else:
    #     logging_config['formatters']['default']['format'] = settings['log.format']

    dictConfig(logging_config)

    root_level = "DEBUG"
    logging.getLogger().setLevel(root_level)

    if isinstance(debug, string_types):
        debug = [debug]
    for package in debug:
        logging.getLogger(package).setLevel(logging.DEBUG)



from typing import Dict, Any, Optional, List, Set

from pydantic import BaseModel

from .tagging import Tagger


class PolicyGroup(BaseModel):
    name: str
    iam: str


class ColumnTag(BaseModel):
    name: str
    # Set to True if it's a root tag
    hasLeafNodes: bool


class PolicyAuthorization(BaseModel):
    auth: str
    value: str
    iam: str


class PolicyCondition(BaseModel):
    type: str


class GroupCondition(PolicyCondition):
    type: str = "groups"
    group: PolicyGroup
    field: Optional[str] = ""

    def dict(
        self,
        *,
        include: Set[str] = None,
        exclude: Set[str] = None,
        by_alias: bool = False,
        skip_defaults: bool = False,
    ) -> Dict[str, Any]:
        """
        The 'field' key should not exist in the payload if it has no value
        as Immuta's API will reject the request. This is because group condition
        can be created in different contexts, some of which require setting 'field'
        and others where 'field' makes no sense.
        """
        if self.field == "":
            exclude = {"field"}
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
        )


class AuthorizationCondition(PolicyCondition):
    type: str = "authorizations"
    authorization: PolicyAuthorization
    field: str


class PurposeCondition(PolicyCondition):
    type: str = "purposes"
    value: str
    field: str


class PolicyCircumstance(BaseModel):
    operator: str
    type: str


class ColumnTagCircumstance(PolicyCircumstance):
    type: str = "columnTags"
    columnTag: ColumnTag


class TagCircumstance(PolicyCircumstance):
    type: str = "tags"
    tag: ColumnTag


class PolicyExceptions(BaseModel):
    # Must be one of 'and', 'or'
    operator: str
    conditions: List[PolicyCondition]


class PolicyAction(BaseModel):
    type: str
    exceptions: PolicyExceptions
    subscriptionType: str
    allowDiscovery: bool = True


class MaskingAction(PolicyAction):
    type: str = "masking"
    tags: List[ColumnTag]


class GlobalPolicy(BaseModel):
    name: str
    id: Optional[int] = None
    type: str
    template: bool = False
    # Don't know what this is supposed to be.
    # No mention in the docs
    ownerRestrictions: Optional[Any] = None
    circumstances: List[PolicyCircumstance]

    def dict(
        self,
        *,
        include: Set[str] = None,
        exclude: Set[str] = None,
        by_alias: bool = False,
        skip_defaults: bool = False,
    ) -> Dict[str, Any]:
        # Remove keys that should not exist in the payload if they have no values
        exclude = set()
        if not self.id:
            exclude.add("id")
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
        )


class GlobalDataPolicy(GlobalPolicy):
    type: str = "data"
    actions: List[MaskingAction]


class GlobalSubscriptionPolicy(GlobalPolicy):
    name: str
    type: str = "subscription"
    actions: List[PolicyAction]


def make_policy_exceptions(iam_groups: List[str]) -> PolicyExceptions:
    conditions = []
    for group in iam_groups:
        # TODO: Generalize for other IAMs
        conditions.append(GroupCondition(group=PolicyGroup(name=group, iam="okta")))
    # TODO: Support other operators
    return PolicyExceptions(operator="or", conditions=conditions)


def make_policy_circumstance(
    tags: List[str], tagger: Tagger
) -> List[PolicyCircumstance]:
    circumstances = []
    for tag in tags:
        circumstances.append(
            ColumnTagCircumstance(
                operator="or",
                columnTag=ColumnTag(name=tag, hasLeafNodes=tagger.is_root_tag(tag)),
            )
        )
    return circumstances


def make_policy_object_from_json(json_policy: Dict[str, Any]) -> GlobalPolicy:
    circumstances = []
    exceptions: Optional[PolicyExceptions] = None
    for circumstance in json_policy["circumstances"]:
        if circumstance["type"] == "columnTags":
            circumstances.append(ColumnTagCircumstance(**circumstance))
        elif circumstance["type"] == "tags":
            circumstances.append(TagCircumstance(**circumstance))
        else:
            raise TypeError(
                f"Unsupported type for policy circumstance: {circumstance['type']}"
            )

    actions = []
    for action in json_policy["actions"]:
        if action.get("exceptions"):
            conditions = []
            for condition in action["exceptions"]["conditions"]:
                if condition["type"] == "groups":
                    conditions.append(GroupCondition(**condition))
                elif condition["type"] == "authorizations":
                    conditions.append(AuthorizationCondition(**condition))
                else:
                    raise TypeError(
                        f"Unsupported type for policy condition: {condition['type']}"
                    )
            exceptions = PolicyExceptions(
                operator=action["exceptions"]["operator"], conditions=conditions
            )
        else:
            exceptions = None

        action.pop("exceptions")
        if action["type"] == "masking":
            actions.append(MaskingAction(**action, exceptions=exceptions))
        else:
            actions.append(PolicyAction(**action, exceptions=exceptions))

    if json_policy["type"] == "subscription":
        return GlobalSubscriptionPolicy(
            id=json_policy["id"],
            name=json_policy["name"],
            type=json_policy["type"],
            template=json_policy["template"],
            circumstances=circumstances,
            actions=actions,
        )
    elif json_policy["type"] == "data":
        return GlobalDataPolicy(
            id=json_policy["id"],
            name=json_policy["name"],
            type=json_policy["type"],
            template=json_policy["template"],
            circumstances=circumstances,
            actions=actions,
        )
    else:
        raise TypeError(f"Unsupported type for Global policy: {json_policy['type']}")


def make_global_subscription_policy(
    policy_name: str, tags: List[str], iam_groups: List[str], tagger: Tagger
) -> GlobalSubscriptionPolicy:
    actions: List[PolicyAction] = []
    actions.append(
        PolicyAction(
            type="subscription",
            subscriptionType="policy",
            exceptions=make_policy_exceptions(iam_groups=iam_groups),
        )
    )
    return GlobalSubscriptionPolicy(
        name=policy_name,
        circumstances=make_policy_circumstance(tags=tags, tagger=tagger),
        actions=actions,
    )



class Paginator(object):
    """
    Handle immuta client pagination.
    assumptions:
     - The result of the client action returns an object with the list of objects in the key 'hits'
     - The client actions have the parameters `size` and `offset`

    example:
    >>> client = ImmutaClient(**client_config)
    >>> with Paginator(client.get_data_source_list, search_text='test') as paginator:
    ...    for data_source in paginator:
    ...        print(data_source['id'])

    """

    def __init__(self, action, *args, **kwargs):
        self.action = action
        self.action_args = args
        self.action_kwargs = kwargs
        self._size = self.action_kwargs.pop("size", 50)
        self._offset = self.action_kwargs.pop("offset", 0)
        self.has_next_page = True

    def __iter__(self):
        while self.has_next_page:
            result = self.next_page()
            for hit in result["hits"]:
                yield hit

    def next_page(self):
        result = self._execute_action()
        self._offset += self._size
        if len(result["hits"]) < self._size:
            self.has_next_page = False
        return result

    def current_page(self):
        return self._offset / self._size

    def _execute_action(self):
        return self.action(
            *self.action_args,
            **self.action_kwargs,
            size=self._size,
            offset=self._offset
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._offset = 0
        self.has_next_page = False



# coding: utf-8

from setuptools import setup, find_packages  # noqa: H301

NAME = "fh-immuta-utils"


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name=NAME,
    version = '0.0.2',
    description="Flatiron Immuta API",
    author_email="data-tooling@flatiron.com",
    url="https://github.com/flatironhealth/fh-immuta-utils",
    packages=find_packages(),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers = [
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    python_requires='>=3.6',
)



from collections import defaultdict
import logging
import os
import glob
from typing import Any, Dict, List, Iterator, Tuple, TYPE_CHECKING

import yaml

from pydantic import BaseModel

from .data_source import DataSourceColumn

if TYPE_CHECKING:
    from immuta_utils.client import ImmutaClient

LOGGER = logging.getLogger(__name__)


class Tag(BaseModel):
    name: str
    source: str = "curated"


class Tagger(object):
    """ Wrapper around managing tags. """

    def __init__(self, config_root: str) -> None:
        # column_name: [tag1, tag2, ...]
        self.tag_map: Dict[str, List[str]] = {}

        # tag_name: [iam_group_1, iam_group_2, ...]
        self.tag_groups: Dict[str, List[str]] = {}
        self.read_configs(config_root=config_root)

    def read_configs(self, config_root: str) -> None:
        for tag_file in glob.glob(os.path.join(config_root, "tags", "*.yml")):
            logging.debug("Reading tag file: %s", tag_file)
            with open(tag_file) as handle:
                contents = yaml.safe_load(handle)
                self.tag_map = {**self.tag_map, **contents.get("TAG_MAP", {})}
                self.tag_groups = {**self.tag_groups, **contents.get("TAG_GROUPS", {})}

    def get_tags_for_column(self, column_name: str) -> List[str]:
        return self.tag_map.get(column_name, [])

    def is_root_tag(self, tag_to_check: str) -> bool:
        for tag in self.tag_groups:
            if "." in tag and tag.split(".")[0] == tag_to_check:
                return True
        return False

    def get_allowed_groups_for_tag(self, tag: str) -> List[str]:
        """ Returns IAM groups allowed access to view things with given tag """
        return self.tag_groups.get(tag, [])

    def tags_to_make(self) -> Iterator[Tuple[str, List[str]]]:
        """
        Yields a dict of either str or List[str] where the key is a parent tag
        and the value is a list of child tags if any, or the parent tag itself if no children.
        """
        all_tags: Dict[str, List[str]] = defaultdict(list)
        for tag_list in self.tag_map.values():
            for tag in tag_list:
                parent = tag.split(".")[0]
                if tag not in all_tags[parent]:
                    all_tags[parent].append(tag)
        for root_tag, children in all_tags.items():
            if len(children) == 1 and children[0] == root_tag:
                children = []
            yield (root_tag, children)

    @classmethod
    def create_message_body_for_tag_creation(
        cls, root_tag: str, children: List[str]
    ) -> Dict[str, Any]:
        if not children:
            return {
                "tags": [Tag(name=root_tag).dict(by_alias=True, skip_defaults=True)]
            }

        return {
            "rootTag": {"name": root_tag, "deleteHierarchy": False},
            "tags": [
                Tag(name=child).dict(by_alias=True, skip_defaults=True)
                for child in children
            ],
        }

    def make_tags(self, client: "ImmutaClient") -> None:
        LOGGER.debug("Creating tags")
        for (root_tag, children) in self.tags_to_make():
            LOGGER.debug(f"Creating root tag: {root_tag}, children: {children}")
            client.create_tag(
                tag_data=self.create_message_body_for_tag_creation(
                    root_tag=root_tag, children=children
                )
            )

    def enrich_columns_with_tagging(
        self, columns: List[DataSourceColumn]
    ) -> List[DataSourceColumn]:
        """Append column tags to a pre-existing list of columns.

        Returns
        -------
        enriched_columns:
            A copy of the list of data source columns with tags added.

        """
        enriched_columns = []
        for c in columns:
            column = c.copy(deep=True)
            # We need to set the source when adding tags, otherwise the API silently rejects the tags list
            # We don't know what other possible values exist for source, but all our currently made tags
            # are curated.
            column.tags = [
                {"name": tag, "source": "curated"}
                for tag in self.get_tags_for_column(column_name=column.name)
            ]
            enriched_columns.append(column)
        return enriched_columns
